require 'strscan'
require 'iconv'

# RLML contains all rlml[http://www.rlml.org] related classes.
module RLML
  VERSION = "0.3"

  # Exception to be raised on errors during parsing.
  class ParseError < Exception; end

  # Exception to be raised on errors during emitting.
  class WriteError < Exception; end

  # CoreParser is a fast, lowlevel parser to be used as a base for
  # higherlevel parsers.
  #
  # CoreParser does *not*:
  # * validate strictly
  # * obey any of the inner semantics of RLML (this is left to higher levels)
  # * handle encodings, all data is passed as-is.
  # 
  # *Note*:: You probably don't want to use this parser directly,
  #          but via the nicer (and more correct) StreamParser.
  #
  # CoreParser is intented to be subclassed.
  class CoreParser
    def initialize
    end

    # Parse the string in _data_ as RLML and generate events.
    def parse(data)
      scanner = StringScanner.new(data)

      start_document

      if scanner.scan %r:^rlml/(.*)[\n\s]+:
        encoding scanner[1].strip
        scanner.pos -= 1        # Step back for whitespace
      else
        encoding nil
      end
 
      until scanner.empty?
        if scanner.scan /\n?\s*#(.*)/
          comment scanner[1]
          next
        end

        scanner.scan /(\s|\n)+/
        
        if scanner.scan /([\w.]+)(\s|\n)*\{/
          start_tag scanner[1]
        elsif scanner.scan /\[(.*?)\]/m
          comment scanner[1]
        elsif scanner.scan /"(.*?([^\\]|\\\\))"/
          data scanner[1].gsub('\t', "\t").gsub('\"', "\"").
                          gsub('\r', "\r").gsub('\n', "\n").gsub('\\\\', "\\")
        elsif scanner.scan /\}/
          end_tag
        else
          unless scanner.rest == ""
            raise ParseError, "Can't parse: `#{scanner.rest}'"
          end
        end
      end
      
      end_document
    end

    protected

    # CoreParser-API event sent before parsing.
    def start_document; end

    # CoreParser-API event sent to specify the encoding.
    #
    # The RLML header <tt>rlml/utf-8</tt> will generate the event
    # <tt>encoding('utf-8')</tt>.
    def encoding(enc); end

    # CoreParser-API event sent on comments.
    #
    # The RLML comment <tt>[ Just another comment. ]</tt> will generate the event
    # <tt>comment(' Just another comment. ')</tt>.
    def comment(cmt); end

    # CoreParser-API event sent on data.
    #
    # The RLML data <tt>"hack,\n hack!"</tt> will generate the event
    # <tt>data("hack,\n hack")</tt>.
    # (Escape chars are converted.)
    def data(data); end

    # CoreParser-API event sent on the beginning of a tag.
    #
    # The RLML fragment <tt>foo.bar {</tt> will generate the event
    # <tt>start_tag("foo.bar")</tt>.
    def start_tag(tag); end

    # CoreParser-API event sent on the end of a tag.
    #
    # The RLML fragment <tt>}</tt> will generate the event
    # <tt>end_tag()</tt>.
    def end_tag; end

    # CoreParser-API event sent after parsing.
    def end_document; end
  end

  # StreamParser is a convenient API for stream based RLML processing.
  #
  # It validates as good as possible and handles encodings using Iconv.
  #
  # On parsing, it sends events as specified in StreamInterface.
  class StreamParser < CoreParser
    # _callback_ will be sent the StreamParser-API events.
    # See StreamInterface on the specifics of this API.
    def initialize(callback)
      @reciever = callback
      @encoding = nil
    end

    # Parse the string in _data_ as RLML and generate events to the
    # callback.
    def parse(data)
      super EncodedString.new(data, @encoding)
    end

    protected

    def start_document
      @tags = []
      @encoding = nil
      @depth = 0
      @reciever.send :start_document
    end

    def end_document
      raise ParseError, "Invalid nesting"  unless @depth == 0
      @reciever.send :end_document
    end

    def encoding(encoding)
      if encoding
        @encoding = encoding
        @reciever.send :encoding, encoding
      else
        # Assume utf-8 and warn?
        raise ParseError, "No encoding given"
      end
    end

    def comment(comment)
      @reciever.send :comment, comment
    end

    def start_tag(tag)
      unless tag =~ /^[A-Za-z_]\w*(.?[A-Za-z_]\w*)*$/ && tag.size < 1024
        raise ParseError, "Invalid tag: `#{tag}'"
      end

      @tags << tag
      @depth += 1
      @reciever.send :start_tag, tag
    end

    def data(data)
      raise ParseError, "Data outside a block is not allowed"  if @depth <= 0
      @reciever.send :data, EncodedString.new(data, @encoding)
    end

    def end_tag
      raise ParseError, "Invalid nesting"  if @depth <= 0
      @depth -= 1      
      @reciever.send :end_tag, @tags.pop
    end
  end

  # EncodedString is the class used to store strings and their encoding.
  # It behaves just like an ordinary string, with two exceptions:
  # 1. You need to pass it's encoding on initialization
  # 2. to_s takes an additional, optional parameter to convert the string
  #    into the given encoding.
  class EncodedString < String
    attr_reader :encoding

    # Create a new EncodedString using the data in _string_ and the
    # _encoding_.
    def initialize(string, encoding)
      super string
      @encoding = encoding
    end

    # Return a copy of the string converted to _encoding_ (if given).
    def to_s(encoding=nil)
      if encoding
        super Iconv.iconv(encoding, @encoding, self).first
      else
        super
      end
    end
  end

  # Tree is the general class containing a RLML TreeParser tree.
  class Tree < Array
    attr_reader :name
    attr_reader :encoding
    
    # Create a new Tree called _name_.
    # You may specify a different encoding, and initial children.
    def initialize(name, encoding='utf-8', *children)
      @name = name.to_s
      @encoding = encoding
      super children
    end

    # Equality---Two Trees are equal if they have the same name,
    # contain the same number of children and if each child is equal
    # to (according to Object#==) the corresponding child in the
    # other array.
    def ==(other)
      self.name == other.name && super
    end
    
    # Returns a string containing a human-readable representation of
    # the tree.
    def inspect
      "#<Tree:" + @name + " " + super + ">"
    end
    
    # Return the children of the tree.  Alternatively, you may as well
    # just handle the tree as if it was an array of children.
    def children
      self
    end

    # Return a string containing the RLML representation of the tree.
    # Generally:
    #   TreeParser.new.parse(tree.to_s) == tree    #=> true
    #
    # If _encoding_ is given, the string is re-encoded as specified.
    def to_s(encoding=nil)
      result = ''
      sw = StreamWriter.new(result)
      sw.encoding encoding || @encoding
      to_s_helper(sw)
      result
    end

    private
    def to_s_helper(sw)
      sw.start_tag(@name)  unless @name == ""
      children.each { |child|
        case child
        when String
          sw.data child
        when Tree
          child.send :to_s_helper, sw
        else
          raise WriteError, "Can't marshal #{child.inspect}"
        end
      }
      sw.end_tag(@name)  unless @name == ""
    end
  end

  # StreamInterface is a mixin with default (empty) event definitions
  # for the RLML::StreamParser-API.
  # 
  # Classes to be used with this interface should include this module
  # and override its methods as needed.
  module StreamInterface
    protected
    
    # Event run before parsing starts.
    def start_document; end

    # Use _encoding_ to encode the RLML data.
    def encoding(encoding); end

    # Tag _name_ started.
    def start_tag(name); end

    # Data _data_ read.
    def data(data); end

    # Comment _comment_ read.
    def comment(comment); end

    # Tag _name_ ended.
    def end_tag(e); end

    # Event run after parsing is finished.
    def end_document; end
  end

  # TreeParser generates a DOM-like Tree.
  class TreeParser
    include StreamInterface

    def initialize
      @parser = StreamParser.new self
      @parents = [Tree.new("")]
    end

    # Parse the string in _data_ as RLML and return the corresponding Tree.
    def parse(data)
      @parser.parse data
      @parents.first
    end

    protected

    def encoding(e)
      @encoding = e
    end

    def start_tag(name)
      tree = Tree.new(name, @encoding)
      @parents.last << tree
      @parents << tree
    end

    def end_tag(name)
      @parents.pop
    end

    def data(data)
      @parents.last << data
    end
  end

  # CoreWriter is a unvalidating, "unsafe", but fast writer.
  # It does no output canonical RLML in all cases.
  # 
  # The output only needs to respond to <tt><<</tt>.
  class CoreWriter
    include StreamInterface

    attr_accessor :output

    def initialize(output)
      @output = output
      @just_after_header = false
    end

    def start_document
      self
    end

    def end_document
      self
    end

    def encoding(e)
      @encoding = e
      @output << "rlml/#{e}"
      self
    end

    def comment(c)
      @output << "[#{c}]"
      self
    end

    def start_tag(name)
      @output << "#{name}{"
      self
    end

    def end_tag
      @output << "}"
      self
    end

    def data(data)
      if data.kind_of? EncodedString
        encoding = data.encoding
      else
        encoding = "utf-8"
      end

      edata = Iconv.iconv(@encoding, encoding, data).first
      
      edata = edata.gsub("\\", '\\\\').gsub("\n", '\n').gsub("\r", '\r').
                    gsub("\"", '\"').gsub("\t", '\t')
          
      @output << '"' << edata << '"'
      self
    end
  end

  # StreamWriter is the preferred way to generate canonical RLML using
  # the StreamInterface-API.
  #
  # StreamWriter always emits valid, canonical RLML.
  #
  # The output only needs to respond to <tt><<</tt>.
  class StreamWriter < CoreWriter
    def initialize(output)
      super
      @depth = 0
      @header_emitted = false
    end

    def emit_required_whitespace
      if @just_after_header
        @output << " "
        @just_after_header = false
      end
    end

    def encoding(e = 'utf-8')
      if e =~ /[\w_-]+/
        @output << "rlml/#{e}"
        @encoding = e
        @just_after_header = true
      else
        raise WriteError, "Invalid encoding: #{e}"
      end
      @header_emitted = true
      self
    end

    def start_tag(name)
      unless name =~ /^[A-Za-z_]\w*(.?[A-Za-z_]\w*)*$/ && name.size < 1024
        raise WriteError, "Invalid name: `#{name}'"
      end
      @depth += 1
      emit_required_whitespace
      super
    end
    
    def end_tag(name = nil)
      raise WriteError, "Invalid nesting"  if @depth <= 0
      # Check against NAME.
      @depth -= 1
      emit_required_whitespace
      super()
    end

    def end_document
      raise WriteError, "No header emitted"  unless @header_emitted
      raise WriteError, "Invalid nesting"  unless @depth == 0
      super
    end
    
    def data(data)
      raise WriteError, "Data outside a block is not allowed"  if @depth <= 0
      emit_required_whitespace
      super
    end
  end
end

# arch-tag: a5a79ce3-fcef-4361-bea1-f9581a482b5c
