require 'test/unit'
require 'rlml'
$VERBOSE = true

class EventLogger
  attr_accessor :events

  def initialize
    @events = []
  end

  def method_missing(*args)
    @events << args
  end
end

class RLMLTest < Test::Unit::TestCase
  def test_stream_parser
    RLML::StreamParser.new(e = EventLogger.new).parse <<EOF
rlml/iso-8859-1
note
{
   to{"Tove"} from{"Jani"} heading{"Reminder"}
   body{"Don't forget me this weekend!"}
}
EOF
   
    assert_equal e.events, \
    [[:start_document], [:encoding, "iso-8859-1"], [:start_tag, "note"],
     [:start_tag, "to"], [:data, "Tove"], [:end_tag, "to"],
     [:start_tag, "from"], [:data, "Jani"], [:end_tag, "from"],
     [:start_tag, "heading"], [:data, "Reminder"], [:end_tag, "heading"],
     [:start_tag, "body"], [:data, "Don't forget me this weekend!"],
     [:end_tag, "body"], [:end_tag, "note"], [:end_document]]


    RLML::StreamParser.new(e = EventLogger.new).parse <<EOF
rlml/utf-8
cd{
    title{"Empire Burlesque"}
    artist{"Bob Dylan"}
    country{"USA"}
    company{"Columbia"}
    price{"10.90"}
    year{"1985"}
}
cd{
    title{"Hide your heart"}
    artist{"Bonnie Tyler"}
    country{"UK"}
    company{"CBS Records"}
    price{"9.90"}
    year{"1988"}
}
EOF

    assert_equal e.events, \
    [[:start_document], [:encoding, "utf-8"], [:start_tag, "cd"],
     [:start_tag, "title"], [:data, "Empire Burlesque"],
     [:end_tag, "title"], [:start_tag, "artist"], [:data, "Bob Dylan"],
     [:end_tag, "artist"], [:start_tag, "country"], [:data, "USA"],
     [:end_tag, "country"], [:start_tag, "company"], [:data, "Columbia"],
     [:end_tag, "company"], [:start_tag, "price"], [:data, "10.90"],
     [:end_tag, "price"], [:start_tag, "year"], [:data, "1985"],
     [:end_tag, "year"], [:end_tag, "cd"], [:start_tag, "cd"],
     [:start_tag, "title"], [:data, "Hide your heart"], [:end_tag, "title"],
     [:start_tag, "artist"], [:data, "Bonnie Tyler"], [:end_tag,
     "artist"], [:start_tag, "country"], [:data, "UK"], [:end_tag,
     "country"], [:start_tag, "company"], [:data, "CBS Records"],
     [:end_tag, "company"], [:start_tag, "price"], [:data, "9.90"],
     [:end_tag, "price"], [:start_tag, "year"], [:data, "1988"],
     [:end_tag, "year"], [:end_tag, "cd"], [:end_document]]
  end

  def test_tree_parser
    tree1 = RLML::TreeParser.new.parse <<EOF
rlml/utf-8
t1{
    "bla"
    t2{"foo"}
    t3{"bar"}
}
EOF

    tree2 = RLML::TreeParser.new.parse %<rlml/utf-8\nt1{"bla"t2{"foo"}t3{"bar"}}>
    tree3 = RLML::TreeParser.new.parse 'rlml/utf-8
t1{
#heheh
"bla"t2{[foo]"foo"[quux]}t3{"bar"
#yay
}}'

    tree4 = RLML::TreeParser.new.parse %<rlml/utf-8\nt1{["dawdaw"\n#]"bla"t2{"foo"}t3{"bar"}}>

    assert_equal tree1, \
    RLML::Tree.new('', nil,
                   RLML::Tree.new("t1", nil,
                                  "bla",
                                  RLML::Tree.new("t2", nil, "foo"),
                                  RLML::Tree.new("t3", nil, "bar")))

    assert_equal tree1, tree2
    assert_equal tree1, tree3
    assert_equal tree1, tree4

    RLML::TreeParser.new.parse <<EOF
rlml/utf-8

database {
   name {"fdfddsfsgd"}
   host {"localhost"}
   user {"ghjughsfjhfs"}
   pass {"sfjnfjfdskjn"}
}

logging {
   path {"/home/foo/logs"}
   level {"verbose"} # options: "verbose"; "errors"; "none"
}

foobars {
   # here's a comment
   extension_path {"/home/foobars/extensions"}
}
EOF
  end

  def test_parser_exceptions
    assert_raises(RLML::ParseError, "No encoding given") {
      RLML::StreamParser.new(EventLogger.new).parse('foo')
    }

    assert_raises(RLML::ParseError, "Can't parse: `&&&'") {
      RLML::StreamParser.new(e = EventLogger.new).parse("rlml/utf-8\n&&&")
    }

    assert_raises(RLML::ParseError, "Invalid tag: `tz.'") {
      RLML::StreamParser.new(e = EventLogger.new).parse("rlml/utf-8\ntz.{}")
    }

    assert_raises(RLML::ParseError, "Invalid tag: `.tz'") {
      RLML::StreamParser.new(e = EventLogger.new).parse("rlml/utf-8\n.tz{}")
    }

    assert_raises(RLML::ParseError, "Invalid tag: `t..z'") {
      RLML::StreamParser.new(e = EventLogger.new).parse("rlml/utf-8\nt..z{}")
    }

    assert_raises(RLML::ParseError,
                  "Invalid tag: `" + "thisistoolong" * 128 + " '") {
      RLML::StreamParser.new(e = EventLogger.new).parse \
        "rlml/utf-8\n" + "thisistoolong" * 128 + "{}"
    }
  
    assert_raises(RLML::ParseError, "Data outside a block is not allowed") {
      RLML::StreamParser.new(e = EventLogger.new).parse %<rlml/utf-8\n"foo">
    }
    
    assert_raises(RLML::ParseError, "Invalid nesting") {
      RLML::StreamParser.new(e = EventLogger.new).parse %<rlml/utf-8\nf{}}>
    }

    assert_raises(RLML::ParseError, "Invalid nesting") {
      RLML::StreamParser.new(e = EventLogger.new).parse %<rlml/utf-8\nf{a{b{}>
    }
  end

  def test_string_escapes
    tree = RLML::TreeParser.new.parse "rlml/utf-8\ntest{\"\\n\\r\\t\\\"\\\\\"}"
    assert_equal tree[0][0], "\n\r\t\"\\"
    assert tree.to_s, "rlml/utf-8\ntest{\"\\n\\r\\t\\\"\\\"}"
  end

  def test_header
    tree1 = RLML::TreeParser.new.parse "rlml/utf-8\nfoo{}"
    tree2 = RLML::TreeParser.new.parse "rlml/utf-8 foo{}"
    tree3 = RLML::TreeParser.new.parse "rlml/utf-8\n#foo\nfoo{}"
    tree4 = RLML::TreeParser.new.parse "rlml/utf-8\n  \n#foo\nfoo{}"

    assert_equal tree1, tree2
    assert_equal tree1, tree3
    assert_equal tree1, tree4
  end

  def test_writer
    # Empty document
    RLML::StreamWriter.new(s='').start_document.encoding.end_document
    assert_equal s, "rlml/utf-8"

    RLML::StreamWriter.new(s='').
      start_document.encoding.
      start_tag('foo').end_tag.
      end_document
    assert_equal s, "rlml/utf-8 foo{}"

    RLML::StreamWriter.new(s='').
      start_document.encoding.
      start_tag('foo').end_tag.
      start_tag('bar').data("yay").end_tag.
      start_tag('baz').data("yay").data("yay").end_tag.
      start_tag('quux').data("yay").data("yay").data("yay").end_tag.
      end_document
    assert_equal s, "rlml/utf-8 foo{}bar{\"yay\"}baz{\"yay\"\"yay\"}quux{\"yay\"\"yay\"\"yay\"}"

    RLML::StreamWriter.new(s='').
      start_document.encoding.
      start_tag('foo').
      start_tag('bar').data("yay").
      start_tag('baz').data("yay").data("yay").
      start_tag('quux').data("yay").data("yay").data("yay").
      end_tag.end_tag.end_tag.end_tag.end_document
    assert_equal s, "rlml/utf-8 foo{bar{\"yay\"baz{\"yay\"\"yay\"quux{\"yay\"\"yay\"\"yay\"}}}}"
  end

  def test_writer_exceptions
    assert_raises(RLML::WriteError, "No header emitted") {
      RLML::StreamWriter.new('').start_document.end_document
    }

    assert_raises(RLML::WriteError, "Invalid nesting") {
      RLML::StreamWriter.new(s='').
        start_document.
        encoding.
        start_tag('foo').
        end_document
    }

    assert_raises(RLML::WriteError, "Invalid nesting") {
      RLML::StreamWriter.new(s='').
        start_document.encoding.
        end_tag('foo').
        end_document
    }

    assert_raises(RLML::WriteError, "Can't marshal 1.0") {
      RLML::Tree.new("foo", nil, 1.0).to_s
    }
  end
end

# arch-tag: ec88a17d-dd1a-43f8-a878-9238b8faaa17
